"""Deferred function evaluation with static and run-time type checks.

https://github.com/martinResearch/postponed/blob/main/README.md
"""
import inspect
import multiprocessing
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Dict, Generic, Iterable, List, TypeVar

from typeguard import check_type

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

P = ParamSpec("P")
R = TypeVar("R")


def check_arguments(
    signature: inspect.Signature, args: P.args, kwargs: Dict[str, Any]  # type: ignore
) -> None:
    # check all required parameters are provided
    all_args = kwargs.copy()
    for name, value in zip(signature.parameters.keys(), args):
        if name in all_args:
            raise ValueError(f"Argument {name} is provided twice")
        all_args[name] = value

    for name in kwargs.keys():
        if name not in signature.parameters:
            raise ValueError(f"Argument {name} not expected")

    for name, v in signature.parameters.items():
        if v.default == inspect._empty:
            if name not in all_args:
                raise ValueError(f"Missing required argument {name}")

    # checking the types using the function check_type from the typeguard library
    for name, value in all_args.items():
        assert name in signature.parameters
        expected_type = signature.parameters[name].annotation
        check_type(value=value, expected_type=expected_type)


class Task(Generic[P, R]):
    """Postponed evaluation of a function with arguments."""

    def __init__(self, f: Callable[P, R], args: P.args, kwargs: P.kwargs):  # type: ignore
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> R:
        return self.f(*self.args, **self.kwargs)


class Postponed(Generic[P, R]):
    """Postponed version of a given function."""

    def __init__(self, f: Callable[P, R], check_inputs: bool = True):
        self._f = f
        self.check_inputs = check_inputs
        if self.check_inputs:
            self._signature = inspect.signature(f)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Callable[[], R]:
        if self.check_inputs:
            check_arguments(signature=self._signature, args=args, kwargs=kwargs)
        return Task(self._f, args, kwargs).__call__


def postponed(
    f: Callable[P, R], /, check_inputs: bool = True
) -> Callable[P, Callable[[], R]]:
    """Lazy version of a given function.

    If `check_inputs` is True then the input arguments and their types are checked as early as possible to fail fast.

    Example:
    >>> p = postponed(print)(5)
    >>> p()
        5
    """
    return Postponed(f, check_inputs=check_inputs).__call__


def execute_tasks_threads(
    tasks: Iterable[Callable[[], R]], max_workers: int
) -> List[R]:
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(task) for task in tasks]
        results = [future.result() for future in futures]
    return results


def execute_tasks_processes(
    tasks: Iterable[Callable[[], R]], max_workers: int
) -> List[R]:
    pool = multiprocessing.Pool(processes=max_workers)
    futures = [pool.apply_async(task, ()) for task in tasks]
    results = [future.get() for future in futures]
    return results



"""Test for the postponed module"""
import multiprocessing

from postponed import execute_tasks_processes, execute_tasks_threads, postponed


def multiply(a: float, b: float) -> float:
    return a * b


def repeat_string(s: str, n: int) -> str:
    return s * n


def tests() -> None:
    postponed_multiply = postponed(multiply)
    tasks = [postponed_multiply(a=1.0, b=2.0), postponed_multiply(a=3.0, b=4.0)]

    max_workers = 3
    pool = multiprocessing.Pool(processes=max_workers)
    futures = [pool.apply_async(task, ()) for task in tasks]
    results = [future.get() for future in futures]
    assert results == [2.0, 12.0]

    results = execute_tasks_processes(tasks=tasks, max_workers=max_workers)
    assert results == [2.0, 12.0]

    results = execute_tasks_threads(tasks=tasks, max_workers=max_workers)
    assert results == [2.0, 12.0]


def test_repeat_string() -> None:
    max_workers = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(processes=max_workers)
    arguments = [("a", 2), ("b", 3)]
    outputs = pool.starmap(repeat_string, arguments)
    print(outputs)
    assert outputs == ["aa", "bbb"]

    # mypy does raise an error when the first argument
    # is a int and not a string in each tuple
    # arguments = [(2, 2), (2, 3)]
    # outputs = pool.starmap(repeat_string, arguments)
    # print(outputs)

    postponed_repeat_string = postponed(repeat_string, check_inputs=True)
    tasks = [
        postponed_repeat_string(s="a", n=2),
        postponed_repeat_string(s="b", n=3),
        postponed_repeat_string(s="c", n=4),
    ]
    outputs = execute_tasks_processes(tasks=tasks, max_workers=max_workers)
    assert outputs == ["aa", "bbb", "cccc"]


if __name__ == "__main__":
    tests()



"""Setup script"""

from setuptools import find_packages, setup

libname = "postponed"

setup(
    name=libname,
    version="0.0.1",
    author="Martin de La Gorce",
    author_email="martin.delagorce@gmail.com",
    description="Module to postpone the execution of a function.",
    packages=find_packages(),
    license="MIT",
    ext_modules=[],  # additional source file(s)),
    include_dirs=[],
    package_data={},
    install_requires=["typeguard"],
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
)



"""Postpone module.

Martin de La Gorce. 2023.
"""
__version__ = "0.0.1"

__all__ = [
    "postponed",
    "execute_tasks_threads",
    "execute_tasks_processes",
    "Postponed",
    "Task",
    "check_arguments",
]

from .postponed import (Task, check_arguments, execute_tasks_processes,
                        execute_tasks_threads, postponed)
